# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, List, Optional, ClassVar
from deprecated.sphinx import deprecated


class DFMetaclass(type):
    def __getattr__(Self, key: str):
        if df := Self.factory.get(key):
            return df
        raise AttributeError(
            "Unknown DesignFormat or Type[DesignFormat] attribute",
            key,
            Self,
        )


@dataclass
class DesignFormat(metaclass=DFMetaclass):
    """
    Metadata about the various possible text or binary representations (views)
    of any design.

    For example, ``DesignFormat.nl`` has the metadata for Netlist views.

    :param id: A lowercase alphanumeric/underscore identifier for the design
        format.
    :param extension: The file extension for designs saved in this format.
    :param full_name: A human-readable name for this design format.
    :param alts: A list of alternate ids used to access the DesignFormat by
        the subscript operator. Includes its OpenLane <3.0.0 enumeration name
        for limited backwards compatibility.
    :param folder_override: The subdirectory when
        :meth:`openlane.state.State.save_snapshot` is called on a state. If
        unset, the value for ``id`` will be used.
    :param multiple: Whether this view may have multiple files (typically, files
        that are different across multiple corners or similar.)
    """

    id: str
    extension: str
    full_name: str
    alts: List[str] = field(default_factory=list)
    folder_override: Optional[str] = None
    multiple: bool = False

    @property
    def folder(self) -> str:
        return self.folder_override or self.id

    @property
    @deprecated(
        "The DesignFormat is directly returned now, no need for .value",
        version="3.0.0",
        action="once",
    )
    def value(self) -> DesignFormat:
        return self

    @property
    @deprecated(
        ".name has been removed because it's redundant, use .id",
        version="3.0.0",
        action="once",
    )
    def name(self) -> str:
        return self.id

    def register(self):
        self.__class__.factory.register(self)

    def __str__(self) -> str:
        return self.id

    def __hash__(self):
        return hash(self.id)

    @staticmethod
    @deprecated(
        "Use DesignFormat.factory.get",
        version="3.0.0",
        action="once",
    )
    def by_id(id: str) -> Optional["DesignFormat"]:
        return DesignFormat.factory.get(id)

    class DesignFormatFactory(object):
        """
        A factory singleton for DesignFormats, allowing them to be registered
        and then retrieved by a string name.

        See https://en.wikipedia.org/wiki/Factory_(object-oriented_programming) for
        a primer.
        """

        _registry: ClassVar[Dict[str, DesignFormat]] = {}

        @classmethod
        def register(Self, df: DesignFormat) -> DesignFormat:
            """
            Adds a DesignFormat to the registry using its
            :attr:`DesignFormat.id`, :attr:`DesignFormat.name` and
            :attr:`DesignFormat.alts` attributes.
            """
            Self._registry[df.id] = df
            for alt in df.alts:
                Self._registry[alt] = df
            return df

        @classmethod
        def get(Self, name: str) -> Optional[DesignFormat]:
            """
            Retrieves a DesignFormat type from the registry using a lookup
            string.

            :param name: The registered name of the Step. Case-insensitive.
            """
            return Self._registry.get(name)

        @classmethod
        def list(Self) -> List[str]:
            """
            :returns: A list of IDs of all registered DesignFormat.
            """
            return [cls.id for cls in Self._registry.values()]

    factory: ClassVar = DesignFormatFactory


# Common Design Formats
DesignFormat(
    "nl",
    "nl.v",
    "Verilog Netlist",
    alts=["NETLIST"],
).register()

DesignFormat(
    "pnl",
    "pnl.v",
    "Powered Verilog Netlist",
    alts=["POWERED_NETLIST"],
).register()

DesignFormat(
    "sdf_pnl",
    "sdf_pnl.v",
    "Powered Verilog Netlist for SDF Simulation (No Fills)",
    alts=["SDF_FRIENDLY_POWERED_NETLIST"],
    folder_override="pnl",
).register()

DesignFormat(
    "logical_pnl",
    "logical_pnl.v",
    "Logical cell-only Powered Verilog Netlist",
    alts=["LOGICAL_POWERED_NETLIST"],
    folder_override="pnl",
).register()

DesignFormat(
    "def",
    "def",
    "Design Exchange Format",
    alts=["def_", "DEF"],
).register()

DesignFormat(
    "lef",
    "lef",
    "Library Exchange Format",
    alts=["LEF"],
).register()

DesignFormat(
    "sdc",
    "sdc",
    "Design Constraints",
    alts=["SDC"],
).register()

DesignFormat(
    "sdf",
    "sdf",
    "Standard Delay Format",
    alts=["SDF"],
    multiple=True,
).register()

DesignFormat(
    "spef",
    "spef",
    "Standard Parasitics Extraction Format",
    alts=["SPEF"],
    multiple=True,  # nom, min, max, ...
).register()

DesignFormat(
    "lib",
    "lib",
    "LIB Timing Library Format",
    alts=["LIB"],
    multiple=True,
).register()

DesignFormat(
    "spice",
    "spice",
    "Simulation Program with Integrated Circuit Emphasis",
    alts=["SPICE"],
).register()

DesignFormat(
    "gds",
    "gds",
    "GDSII Stream",
    alts=["GDS"],
).register()

DesignFormat(
    "vh",
    "vh",
    "Verilog Header",
    alts=["VERILOG_HEADER"],
).register()



#!/usr/bin/env python3
# Copyright 2025 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from reader import click_odb, click, odb
import grt as GRT


def average_location(instances):
    locations = [instance.getLocation() for instance in instances]
    x_sum = sum(loc[0] for loc in locations)
    y_sum = sum(loc[1] for loc in locations)
    return (x_sum // len(locations), y_sum // len(locations))


@click.command()
@click_odb
def cli(reader):
    grt = reader.design.getGlobalRouter()
    dpl = reader.design.getOpendp()

    insts_to_temporarily_lock_then_unlock_later = []
    for inst in reader.block.getInsts():
        if inst.getPlacementStatus() != "LOCKED":
            insts_to_temporarily_lock_then_unlock_later.append(
                (inst, inst.getPlacementStatus())
            )
            inst.setPlacementStatus("LOCKED")

    reader._grt_setup(grt)

    grt_inc = GRT.IncrementalGRoute(grt, reader.block)
    i = 0

    for target_info in reader.config["INSERT_ECO_BUFFERS"]:
        target_name, target_pin = target_info["target"].split("/")
        name_escaped = reader.escape_verilog_name(target_name)
        buffer_master = target_info["buffer"]

        master = reader.db.findMaster(buffer_master)
        if master is None:
            print(
                f"[ERROR] Buffer type '{buffer_master}' not found.",
                file=sys.stderr,
            )
            exit(-1)

        target = reader.block.findInst(name_escaped)
        if target is None:
            print(f"[ERROR] Instance '{target_name}' not found.", file=sys.stderr)
            exit(-1)

        target_iterm = target.findITerm(target_pin)
        if target_iterm is None:
            print(
                f"[ERROR] Pin '{target_pin}' not found for instance '{target_name}'.",
                file=sys.stderr,
            )
            exit(-1)

        net = target_iterm.getNet()
        if net is None:
            print(
                f"[ERROR] Net not found on pin '{target_pin}' of instance '{target_name}'.",
                file=sys.stderr,
            )
            exit(-1)

        new_buf_name = f"eco_buffer_{i}"
        new_net_name = f"eco_buffer_{i}_net"
        while (
            reader.block.findInst(new_buf_name) is not None
            or reader.block.findNet(new_net_name) is not None
        ):
            i += 1
            new_buf_name = f"eco_buffer_{i}"
            new_net_name = f"eco_buffer_{i}_net"

        # Prepare buffer cell, net
        eco_buffer = odb.dbInst.create(reader.block, master, new_buf_name)
        eco_net = odb.dbNet.create(reader.block, new_net_name)
        buffer_iterms = eco_buffer.getITerms()
        buffer_a = None
        for iterm in buffer_iterms:
            if iterm.isInputSignal():
                buffer_a = iterm
                break  # Exit loop once input is found
        if buffer_a is None:
            print(
                f"[ERROR] Buffer {buffer_master} has no input signals.",
                file=sys.stderr,
            )
            exit(-1)

        buffer_x = None
        for iterm in buffer_iterms:
            if iterm.isOutputSignal():
                buffer_x = iterm
                break  # Exit loop once output is found
        if buffer_x is None:
            print(
                f"[ERROR] Buffer {buffer_master} has no output signals.",
                file=sys.stderr,
            )
            exit(-1)

        location_instances = [target]
        net_iterms = net.getITerms()
        if target_iterm.getIoType() == "INPUT":
            driver_iterms = [
                iterm for iterm in net_iterms if iterm.getIoType() in ["OUTPUT"]
            ]
            drivers = [iterm.getInst() for iterm in driver_iterms]
            location_instances.extend(drivers)

            target_iterm.disconnect()
            buffer_a.connect(net)
            buffer_x.connect(eco_net)
            target_iterm.connect(eco_net)
        elif target_iterm.getIoType() == "OUTPUT":
            sink_iterms = [
                iterm for iterm in net_iterms if iterm.getIoType() in ["INPUT", "INOUT"]
            ]
            sinks = [iterm.getInst() for iterm in sink_iterms]
            location_instances.extend(sinks)

            target_iterm.disconnect()
            target_iterm.connect(eco_net)
            buffer_a.connect(eco_net)
            buffer_x.connect(net)
        else:
            print(
                f"[ERROR] {target_name}/{target_pin} is neither an INPUT or an OUTPUT and is unsupported by this script. To buffer an INOUT port, buffer its drivers instead.",
                file=sys.stderr,
            )
            exit(-1)

        if target_info.get("placement") is not None:
            eco_x, eco_y = target_info["placement"]
            eco_x = reader.block.micronsToDbu(float(eco_x))
            eco_y = reader.block.micronsToDbu(float(eco_y))
            eco_loc = (eco_x, eco_y)
        else:
            eco_loc = average_location(location_instances)

        eco_buffer.setOrient("R0")
        eco_buffer.setLocation(*eco_loc)
        eco_buffer.setPlacementStatus("PLACED")
        grt.addDirtyNet(net)
        grt.addDirtyNet(eco_net)

    site = reader.rows[0].getSite()
    max_disp_x = int(
        reader.design.micronToDBU(reader.config["PL_MAX_DISPLACEMENT_X"])
        / site.getWidth()
    )
    max_disp_y = int(
        reader.design.micronToDBU(reader.config["PL_MAX_DISPLACEMENT_Y"])
        / site.getHeight()
    )
    dpl.detailedPlacement(max_disp_x, max_disp_y)

    grt_inc.updateRoutes(True)

    for inst, previous_status in insts_to_temporarily_lock_then_unlock_later:
        inst.setPlacementStatus(previous_status)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# Copyright 2025 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from reader import click_odb, click, odb
import grt as GRT


@click.command()
@click_odb
def cli(reader):
    grt = reader.design.getGlobalRouter()
    dpl = reader.design.getOpendp()

    insts_to_temporarily_lock_then_unlock_later = []
    for inst in reader.block.getInsts():
        if inst.getPlacementStatus() != "LOCKED":
            insts_to_temporarily_lock_then_unlock_later.append(
                (inst, inst.getPlacementStatus())
            )
            inst.setPlacementStatus("LOCKED")

    reader._grt_setup(grt)

    diode_master, diode_pin = reader.config["DIODE_CELL"].split("/")

    # print(grt)
    grt_inc = GRT.IncrementalGRoute(grt, reader.block)
    i = 0
    for target_info in reader.config["INSERT_ECO_DIODES"]:
        target_name, target_pin = target_info["target"].split("/")
        name_escaped = reader.escape_verilog_name(target_name)

        target = reader.block.findInst(name_escaped)
        if target is None:
            print(
                f"[ERROR] Instance '{target_name}' not found.",
                file=sys.stderr,
            )
            exit(-1)

        master = reader.db.findMaster(diode_master)
        if master is None:
            print(
                f"[ERROR] Cell kind '{diode_master}' not found.",
                file=sys.stderr,
            )
            exit(-1)

        target_iterm = target.findITerm(target_pin)
        if target_iterm is None:
            print(
                f"[ERROR] Pin '{target_pin}' not found for instance {target_name}.",
                file=sys.stderr,
            )
            exit(-1)

        if target_iterm.getIoType() not in ["INPUT", "INOUT"]:
            print(
                f"[ERROR] Pin {target_info['target']} is an OUTPUT pin.",
                file=sys.stderr,
            )
            exit(-1)

        net = target_iterm.getNet()
        if net is None:
            print(
                f"[ERROR] Pin {target_info['target']} has no nets connected.",
                file=sys.stderr,
            )
            exit(-1)

        eco_diode_name = f"eco_diode_{i}"
        while reader.block.findInst(eco_diode_name) is not None:
            i += 1
            eco_diode_name = f"eco_diode_{i}"

        eco_diode = odb.dbInst.create(
            reader.block,
            master,
            eco_diode_name,
        )

        diode_iterm = eco_diode.findITerm(diode_pin)
        if diode_iterm is None:
            print(
                f"[ERROR] Pin '{diode_pin}' on ECO diode not found- invalid DIODE_CELL definition.",
                file=sys.stderr,
            )
            exit(-1)

        sys.stdout.flush()

        if target_info["placement"] is not None:
            x, y = target_info["placement"]
            x = reader.block.micronsToDbu(float(x))
            y = reader.block.micronsToDbu(float(y))
        else:
            x, y = target.getLocation()

        eco_diode.setOrient("R0")
        eco_diode.setLocation(x, y)
        eco_diode.setPlacementStatus("PLACED")

        diode_iterm.connect(net)
        grt.addDirtyNet(net)

    site = reader.rows[0].getSite()
    max_disp_x = int(
        reader.design.micronToDBU(reader.config["PL_MAX_DISPLACEMENT_X"])
        / site.getWidth()
    )
    max_disp_y = int(
        reader.design.micronToDBU(reader.config["PL_MAX_DISPLACEMENT_Y"])
        / site.getHeight()
    )
    dpl.detailedPlacement(max_disp_x, max_disp_y)

    grt_inc.updateRoutes(True)

    for inst, previous_status in insts_to_temporarily_lock_then_unlock_later:
        inst.setPlacementStatus(previous_status)

    reader.design.writeDef("out.def")


if __name__ == "__main__":
    cli()
